/*
    Copyright 2005,2006,2007,2008,2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <direct.h>
    #include <windows.h>
#else
    #include <dirent.h>

    #define strnicmp    strncasecmp
    #define stricmp     strcasecmp
    #define stristr     strcasestr
#endif

typedef uint8_t     u8;



#define VER         "0.1.2b"
#define CHKHELP(x)  !stricmp(p, x)
#define TOKSZ       64      // input string
#define MAXDIZ      200     // do you need more?
#define MAXROW      23      // classical 80x25
#define WBEXT       ".wb"
#define WB1         wb[i].wb1
#define WB2         wb[i].wb2
#define WB1SZ       31
#define WB2SZ       53



typedef struct {
    u8      *wb1;
    u8      *wb2;
} wb_t;



int fgetz(u8 *buff, int size, FILE *fd);
void destroy_wb_db(wb_t *wb);
wb_t *create_wb_db(int num);
int check_wildcard(u8 *fname, u8 *wildcard);
void goto_program_folder(u8 *argv0);
void my_dir(void);
int found(u8 *name);
int help(void);
FILE *load(int num);
int get_terms(u8 *name);
void std_err(u8 *error);



int     maxdiz;
u8      dizmini[MAXDIZ][16],    // mini name
        dizname[MAXDIZ][64];    // full name



int main(int argc, char *argv[]) {
    wb_t    *wb;
    int     i,
            row;
    u8      token[TOKSZ + 1],
            ans[16],
            *p;

    setbuf(stdout, NULL);

    fputs("\n"
        "CmdDiz "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n"
        "Dictionaries: http://www.freelang.net/dictionary/index.html#list_alpha\n"
        "\n", stdout);

    wb = NULL;
    maxdiz = 0;
    if(argc > 1) {
        found(argv[1]);
    } else {
        dizname[0][0] = 0;
        goto_program_folder(argv[0]);
        printf("- search available dictionaries:\n");
        my_dir();
    }
    wb = create_wb_db(help());

    for(;;) {
        printf("\n: ");
        if(fgetz(token, TOKSZ, stdin) < 0) break;

        if(*token == '/') { // command
            p = token + 1;

            for(i = 0; i < maxdiz; i++) {   // check new dictionary to load
                if(CHKHELP(dizmini[i])) {
                    destroy_wb_db(wb);
                    wb = create_wb_db(i);
                    break;
                }
            }
            if(i < maxdiz) continue;

            if(CHKHELP("?") || CHKHELP("help")) {
                help();
                continue;
            } else if(CHKHELP("quit") || CHKHELP("exit")) {
                break;
            }

            printf("- unexistent command\n");
            continue;
        }

        row = 0;
        for(i = 0; WB1 && WB2; i++) {
            if((check_wildcard(WB1, token) < 0) && (check_wildcard(WB2, token) < 0)) continue;
            if(row == MAXROW) {
                printf("...");
                if(fgetz(ans, sizeof(ans), stdin) < 0) break;
                if(tolower(ans[0]) == 'q') break;
                row = 0;
            }
            printf("%-31s%s\n", WB1, WB2);
            row++;
        }
    }

    destroy_wb_db(wb);
    return(0);
}



int fgetz(u8 *buff, int size, FILE *fd) {
    u8      *p;

    fflush(fd);
    if(!fgets(buff, size, fd)) return(-1);
    for(p = buff; *p && (*p != '\n') && (*p != '\r'); p++);
    *p = 0;
    return(p - buff);
}



void destroy_wb_db(wb_t *wb) {
    int     i;

    if(!wb) return;
    for(i = 0; WB1 && WB2; i++) {
        free(WB1);
        free(WB2);
    }
    free(wb);
}



wb_t *create_wb_db(int num) {
    struct  stat    xstat;
    FILE    *fd;
    wb_t    *wb;
    int     i,
            tot;
    u8      wb1[WB1SZ + 1],
            wb2[WB2SZ + 1];

    fd = load(num);

    fstat(fileno(fd), &xstat);
    tot = (xstat.st_size / (WB1SZ + WB2SZ)) + 1;

    wb = malloc(sizeof(wb_t) * tot);
    if(!wb) std_err(NULL);
    wb1[WB1SZ] = 0;
    wb2[WB2SZ] = 0;

    for(i = 0; fread(wb1, WB1SZ, 1, fd) && fread(wb2, WB2SZ, 1, fd); i++) {
        WB1 = strdup(wb1);
        WB2 = strdup(wb2);
    }
    WB1 = NULL;
    WB2 = NULL;

    fclose(fd);
    return(wb);
}



int check_wildcard(u8 *fname, u8 *wildcard) {
    u8      *f,
            *w,
            *a;

    f = fname;
    w = wildcard;
    a = NULL;
    while(*f || *w) {
        if(!*w && !a) return(-1);
        if(*w == '?') {
            if(!*f) break;
            w++;
            f++;
        } else if(*w == '*') {
            w++;
            a = w;
        } else {
            if(!*f) break;
            if(tolower(*f) != tolower(*w)) {
                if(!a) return(-1);
                f++;
                w = a;
            } else {
                f++;
                w++;
            }
        }
    }
    if(*f || *w) return(-1);
    return(0);
}



void goto_program_folder(u8 *argv0) {
    u8      path[2000],
            *p;

#ifdef WIN32
    GetModuleFileName(NULL, path, sizeof(path));
#else
    sprintf(path, "%.*s", sizeof(path), argv0);
#endif
    p = strrchr(path, '\\');
    if(!p) p = strrchr(path, '/');
    if(p) *p = 0;
    chdir(path);
}



void my_dir(void) {
    u8      *p;

#ifdef WIN32

    static int      winnt = -1;
    OSVERSIONINFO   osver;
    WIN32_FIND_DATA wfd;
    HANDLE  hFind;

    if(winnt < 0) {
        osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
        GetVersionEx(&osver);
        if(osver.dwPlatformId >= VER_PLATFORM_WIN32_NT) {
            winnt = 1;
        } else {
            winnt = 0;
        }
    }

    if(winnt) { // required to avoid problems with Vista and Windows7!
        hFind = FindFirstFileEx("*.*", FindExInfoStandard, &wfd, FindExSearchNameMatch, NULL, 0);
    } else {
        hFind = FindFirstFile("*.*", &wfd);
    }
    if(hFind == INVALID_HANDLE_VALUE) return;

    while(FindNextFile(hFind, &wfd)) {
        if(!strcmp(wfd.cFileName, ".") || !strcmp(wfd.cFileName, "..")) continue;

        if(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;

        p = strrchr(wfd.cFileName, '.');
        if(!p || stricmp(p, WBEXT)) continue;

        if(found(wfd.cFileName) < 0) break;
    }
    FindClose(hFind);

#else

    struct  stat    xstat;
    struct  dirent  **namelist;
    int             n,
                    i;

    n = scandir(".", &namelist, NULL, NULL);
    if(n >= 0) {
        for(i = 2; i < n; i++) {
            if(stat(namelist[i]->d_name, &xstat) < 0) std_err(NULL);
            if(S_ISDIR(xstat.st_mode)) continue;

            p = strrchr(namelist[i]->d_name, '.');
            if(!p || stricmp(p, WBEXT)) continue;

            if(found(namelist[i]->d_name) < 0) {
                while(i < n) free(namelist[i++]);
                break;
            }

            free(namelist[i]);
        }
        free(namelist);
    }

#endif
}



int found(u8 *name) {
    int     i,
            same;
    u8      *p;

    if(!stricmp(name, "--help") || !stricmp(name, "/h") || !stricmp(name, "/?") || !stricmp(name, "-h")) {
        std_err(
            "launch this tool without arguments to load the dictionaries in the\n"
            "       program's folder or specify the name of the dictionary file\n");
        return(-1);
    }

    if(maxdiz == MAXDIZ) {
        printf("\nAlert: has been reached the maximum number of dictionaries\n");
        return(-1);
    }

    printf("  %s\n", name);
    sprintf(dizname[maxdiz], "%.*s", sizeof(dizname[maxdiz]), name);

    p = strchr(name, '_');
    if(p) *p = 0;

    p = strchr(name, ' ');
    if(p) {
        memmove(name + 3, p + 1, 3);
        name[6] = 0;
    } else {
        name[3] = 0;
    }

    memset(dizmini[maxdiz], 0, sizeof(dizmini[maxdiz]));
    for(p = dizmini[maxdiz]; *name; p++, name++) {
        *p = tolower(*name);
    }

    same = 1;
    do {
        for(i = 0; i < maxdiz; i++) {  // check for duplicated names
            if(!strcmp(dizmini[i], dizmini[maxdiz])) {
                sprintf(p, "%d", same);
                same++;
            }
        }
    } while(i < maxdiz);

    maxdiz++;
    return(0);
}



int help(void) {
    int     i,
            def = 0;

    printf("\n"
        "- Commands:\n"
        "  /?    or /help = help\n"
        "  /quit or /exit = exit\n"
        "\n"
        "- Languages:\n");
    for(i = 0; i < maxdiz; i++) {
        printf("  /%-10s = %s   %d\n", dizmini[i], dizname[i], get_terms(dizname[i]));
        if(!stricmp(dizname[i], "english")) def = i;
    }
    return(def);
}



int get_terms(u8 *name) {
    struct  stat    xstat;

    if(stat(name, &xstat) < 0) std_err(NULL);
    return(xstat.st_size / (WB1SZ + WB2SZ));
}



FILE *load(int num) {
    FILE    *fd;

    if(!dizname[num][0]) std_err("no dictionary available in the program's folder");
    printf("\n- load dictionary: %s\n", dizname[num]);
    fd = fopen(dizname[num], "rb");
    if(!fd) std_err(NULL);
    return(fd);
}



void std_err(u8 *error) {
    u8      ans[16];

    if(error) {
        fprintf(stderr, "\nError: %s\n", error);
    } else {
        perror("\nError");
    }
    printf("\nPress RETURN to quit\n");
    fgets(ans, sizeof(ans), stdin);
    exit(1);
}


