/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

/*
  thanx to Andy Short of the Block Adverts project (http://hostsfile.mine.nu/) for the ideas and the wide testing
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#include <ctype.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define in_addr_t   uint32_t
    #define close   closesocket
    #define sleep   Sleep
    #define sleepms sleep
    #define ONESEC  1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define sleepms(X)  usleep(X * 1000)
    #define ONESEC  1
#endif

#ifdef WIN32
    #define quick_thread(NAME, ARG) DWORD WINAPI NAME(ARG)
    #define thread_id   HANDLE
#else
    #define quick_thread(NAME, ARG) void *NAME(ARG)
    #define thread_id   pthread_t
#endif

thread_id quick_threadx(void *func, void *data) {
    thread_id   tid;
#ifdef WIN32
    DWORD   tmp;

    tid = CreateThread(NULL, 0, func, data, 0, &tmp);
    if(!tid) return(0);
#else
    if(pthread_create(&tid, NULL, func, data)) return(0);
#endif
    return(tid);
}

void quick_threadz(thread_id tid) {
#ifdef WIN32
    DWORD   ret;

    for(;;) {
        GetExitCodeThread(tid, &ret);
        if(!ret) break;
        sleep(100);
    }
#else
    pthread_join(tid, NULL);
#endif
}

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER             "0.1"
#define UNEXISTENT      "none.unexistent"

/* Length of DNS header. */
#define DNS_HDRLEN      12

/* type values  */
#define T_A             1           /* host address */
#define T_NS            2           /* authoritative name server */
#define T_MD            3           /* mail destination (obsolete) */
#define T_MF            4           /* mail forwarder (obsolete) */
#define T_CNAME         5           /* canonical name */
#define T_SOA           6           /* start of authority zone */
#define T_MB            7           /* mailbox domain name (experimental) */
#define T_MG            8           /* mail group member (experimental) */
#define T_MR            9           /* mail rename domain name (experimental) */
#define T_NULL          10          /* null RR (experimental) */
#define T_WKS           11          /* well known service */
#define T_PTR           12          /* domain name pointer */
#define T_HINFO         13          /* host information */
#define T_MINFO         14          /* mailbox or mail list information */
#define T_MX            15          /* mail routing information */
#define T_TXT           16          /* text strings */
#define T_RP            17          /* responsible person (RFC 1183) */
#define T_AFSDB         18          /* AFS data base location (RFC 1183) */
#define T_X25           19          /* X.25 address (RFC 1183) */
#define T_ISDN          20          /* ISDN address (RFC 1183) */
#define T_RT            21          /* route-through (RFC 1183) */
#define T_NSAP          22          /* OSI NSAP (RFC 1706) */
#define T_NSAP_PTR      23          /* PTR equivalent for OSI NSAP (RFC 1348 - obsolete) */
#define T_SIG           24          /* digital signature (RFC 2535) */
#define T_KEY           25          /* public key (RFC 2535) */
#define T_PX            26          /* pointer to X.400/RFC822 mapping info (RFC 1664) */
#define T_GPOS          27          /* geographical position (RFC 1712) */
#define T_AAAA          28          /* IPv6 address (RFC 1886) */
#define T_LOC           29          /* geographical location (RFC 1876) */
#define T_NXT           30          /* "next" name (RFC 2535) */
#define T_EID           31          /* ??? (Nimrod?) */
#define T_NIMLOC        32          /* ??? (Nimrod?) */
#define T_SRV           33          /* service location (RFC 2052) */
#define T_ATMA          34          /* ??? */
#define T_NAPTR         35          /* naming authority pointer (RFC 2168) */
#define T_KX            36          /* Key Exchange (RFC 2230) */
#define T_CERT          37          /* Certificate (RFC 2538) */
#define T_A6            38          /* IPv6 address with indirection (RFC 2874) */
#define T_DNAME         39          /* Non-terminal DNS name redirection (RFC 2672) */
#define T_OPT           41          /* OPT pseudo-RR (RFC 2671) */
#define T_DS            43          /* Delegation Signature(RFC 3658) */
#define T_IPSECKEY      45          /* draft-ietf-ipseckey-rr */
#define T_RRSIG         46          /* future RFC 2535bis */
#define T_NSEC          47          /* future RFC 2535bis */
#define T_DNSKEY        48          /* future RFC 2535bis */
#define T_TKEY          249         /* Transaction Key (RFC 2930) */
#define T_TSIG          250         /* Transaction Signature (RFC 2845) */
#define T_WINS          65281       /* Microsoft's WINS RR */
#define T_WINS_R        65282       /* Microsoft's WINS-R RR */

/* Class values */
#define C_IN            1           /* the Internet */
#define C_CS            2           /* CSNET (obsolete) */
#define C_CH            3           /* CHAOS */
#define C_HS            4           /* Hesiod */
#define C_NONE          254         /* none */
#define C_ANY           255         /* any */
#define C_FLUSH         (1<<15)     /* High bit is set for MDNS cache flush */



typedef struct {
    u8          *host;
    in_addr_t   hosts_ip;   // backup for restoring the original hosts entry
    u8          comment;
    u8          ok;
} list_t;



void check_overwrite(u8 *fname);
quick_thread(dnscheck, list_t *list);
u8 *get_next(FILE *fd, in_addr_t *hosts_ip);
void put_list(FILE *fd, list_t *list);
void *get_list(FILE *fd, int tot);
int fgetz(FILE *fd, u8 *data, int size);
int putns(u8 *data, u8 *str, int max);
int getxx(uint8_t *data, uint32_t *ret, int bits);
int putxx(uint8_t *data, uint32_t num, int bits);
int timeout(int sock, int secs);
u32 resolv(int sd, int tid, struct sockaddr_in *peer, int query, char *host);
void std_err(void);



struct sockaddr_in  dns_peer;
FILE    *fdlog      = NULL,
        *fddel      = NULL;
int     tabmode     = 0,
        skiplocal   = 0,
        verbose     = 0,
        done        = 0,    // not the max of thread safe but enough
        percval     = 0,
        tothosts    = 0,
        waitms      = 0,
        dns_query   = T_A,
        dns_query2  = 0,
        dns_error   = 0;



int main(int argc, char *argv[]) {
    thread_id   *tid;
    FILE    *fd;
    list_t  **list;
    u32     starttime;
    int     i,
            tmax,
            threads     = 16;
    u8      *dns_ip,
            *input,
            *output,
            *logfile    = NULL,
            *delfile    = NULL;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fputs("\n"
        "hosts file/list DNS checker "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 3) {
        fprintf(stderr, "\n"
            "Usage: %s [options] <DNS_server> <input_hosts_file> <output_hosts_file>\n"
            "\n"
            "Options:\n"
            "-m      enable TAB mode, input_hosts_file is handled as a list of hosts\n"
            "        separated by tabs, spaces, line-feeds (so one per line) or commas\n"
            "-t MS   wait a specific amount of milliseconds between each DNS query (%d)\n"
            "-l FD   exactly as the -v option but all the messages are logged in file FD\n"
            "-d FD   name of the file where logging all the dropped hostnames\n"
            "-s      skip the DNS checking of the hosts that don't have a dot in the name\n"
            "        like for example localhost, xbox, home_computer and so on\n"
            "-v      verbose, show all hosts and IP addresses\n"
            "-x NUM  number of threads to use for making the scanning faster (%d)\n"
            "-q NUM  type of DNS query to perform like 1 for A records (default), 2 for NS,\n"
            "        15 for MX, 5 for CNAME, 6 for SOA and so on (more in the source code)\n"
            "-Q NUM  exactly as above but this one is performed ONLY if the previous one\n"
            "        returns a negative value, for example if record A can't be resolved\n"
            "\n"
            "the output file can have also the same name of the input one in which case\n"
            "will be overwritten (after user's confirmation) when all the process ends\n"
            "\n"
            "Note: in the default \"hosts file\" mode any line containing a # will be not\n"
            "      processed and will copied as is in the output file\n"
            "\n", argv[0], waitms, threads);
        exit(1);
    }

    argc -= 3;
    for(i = 1; i < argc; i++) {
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            fprintf(stderr, "\nError: wrong argument (%s)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case 'm': tabmode       = 1;                break;
            case 't': waitms        = atoi(argv[++i]);  break;
            case 'l': logfile       = argv[++i];        break;
            case 'd': delfile       = argv[++i];        break;
            case 's': skiplocal     = 1;                break;
            case 'v': verbose       = 1;                break;
            case 'x': threads       = atoi(argv[++i]);  break;
            case 'q': dns_query     = atoi(argv[++i]);  break;
            case 'Q': dns_query2    = atoi(argv[++i]);  break;
            default: {
                fprintf(stderr, "\nError: wrong argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }

    dns_ip = argv[argc];
    input  = argv[argc + 1];
    output = argv[argc + 2];

    fdlog = stderr;
    if(logfile) {
        fprintf(stderr, "- create log file %s\n", logfile);
        check_overwrite(logfile);
        fdlog = fopen(logfile, "wb");
        if(!fdlog) std_err();
        verbose = 1;
    }

    fprintf(fdlog, "- open file %s\n", input);
    fd = fopen(input, "rb");
    if(!fd) std_err();

    tothosts = (int)get_list(fd, -1);
    if(!tothosts) {
        fprintf(fdlog, "\n"
            "Error: there are no hosts in your input file\n"
            "       try to change mode or recheck it\n");
        exit(1);
    }
    fprintf(fdlog, "- found %d hosts\n", tothosts);

    dns_peer.sin_addr.s_addr = inet_addr(dns_ip);
    dns_peer.sin_port        = htons(53);
    dns_peer.sin_family      = AF_INET;
    if(dns_peer.sin_addr.s_addr == htonl(INADDR_NONE)) {
        fprintf(fdlog, "\nError: invalid DNS IP address (%s)\n", dns_ip);
        exit(1);
    }

    fprintf(fdlog, "- check if you are using a non-standard DNS, resolv \"%s\"\n", UNEXISTENT);
    if(resolv(0, 1, &dns_peer, T_A, UNEXISTENT) != htonl(INADDR_NONE)) {
        fprintf(fdlog, "\n"
            "Error: you are using a non-standard DNS like OpenDNS or BigPond\n"
            "       change it to a standard one and retry\n");
        exit(1);
    }

    if(threads <= 0) exit(1);
    tid = malloc(sizeof(thread_id) * threads);
    if(!tid) std_err();
    list = malloc(sizeof(list_t *) * threads);
    if(!list) std_err();

    fprintf(fdlog, "- get list of hosts for each thread\n");
    tmax = tothosts / threads;
    rewind(fd);
    list[0] = (list_t *)get_list(fd, tmax + (tothosts % threads));
    for(i = 1; i < threads; i++) {
        list[i] = (list_t *)get_list(fd, tmax);
    }
    fclose(fd);

    percval = tothosts / 100;
    if(!percval) percval = 1;

    fprintf(fdlog, "- start scanning threads, please wait\n");
    starttime = time(NULL);
    for(i = 0; i < threads; i++) {
        tid[i] = quick_threadx(dnscheck, list[i]);
        sleepms(100);   // not needed but it's good
    }

    sleep(ONESEC);
    for(i = 0; i < threads; i++) {
        quick_threadz(tid[i]);
    }
    fprintf(stderr, "\r  %3d%%\n", 100);

    fprintf(fdlog, "\n- scanning finished in %u seconds\n", (u32)(time(NULL) - starttime));

    if(!strcmp(output, "-")) {
        fd = stdout;
    } else {
        fprintf(stderr, "- create file %s\n", output);
        check_overwrite(output);
        fd = fopen(output, "wb");
        if(!fd) std_err();
    }

    if(delfile) {
        fprintf(stderr, "- create file %s\n", delfile);
        check_overwrite(delfile);
        fddel = fopen(delfile, "wb");
        if(!fddel) std_err();
    }

    for(i = 0; i < threads; i++) {
        put_list(fd, list[i]);
    }

    if(fd != stdout) fclose(fd);
    if(fdlog) fclose(fdlog);
    if(fddel) fclose(fddel);
    if(dns_error) {
        fprintf(stderr, "\n"
            "- have been found some DNS errors (errors different than 2 and 3)\n"
            "  is possible that the DNS server you have used has an anti-flooding mechanism\n"
            "  and so banned your IP or doesn't support a particular type of query (in case\n"
            "  you used -q/-Q) or they are another way to say that a specific host can't be\n"
            "  resolved (for example if the DNS uses an internal banning policy for some\n"
            "  malicious hosts and domains) so in this case you should ignore this warning\n");
    }
    return(0);
}



void check_overwrite(u8 *fname) {
    FILE    *fd;
    u8      ans[16];

    fd = fopen(fname, "rb");
    if(!fd) return;
    fclose(fd);
    fprintf(stderr, "  the file already exists, do you want to overwrite it (y/N)? ");
    fgets(ans, sizeof(ans), stdin);
    if(tolower(ans[0]) != 'y') exit(1);
}



quick_thread(dnscheck, list_t *list) {
    in_addr_t   ip,
                ip2;
    int     sd,
            i;

    if(!list[0].host) return(0);
    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    for(i = 0; list[i].host; i++) {
        if(list[i].comment) continue; // preserve comments
        if(!list[i].host[0]) continue;
        if(skiplocal && !strchr(list[i].host, '.')) continue;

        ip = resolv(sd, i, &dns_peer, dns_query, list[i].host);
        if(verbose) {
            fprintf(fdlog, "  %08x %s\n", ip, list[i].host);
        } else {
            fputc('.', stderr);
        }
        if(ip == htonl(INADDR_NONE)) {
            if(dns_query2) {
                ip2 = resolv(sd, i, &dns_peer, dns_query2, list[i].host);
                if(ip2 == htonl(INADDR_NONE)) list[i].ok = 0;    // remove it
            } else {
                list[i].ok = 0;    // remove it
            }
        }
        // else if(ip == htonl(INADDR_LOOPBACK))

        if(!(done % percval)) fprintf(stderr, "\r  %3d%%", (done * 100) / tothosts);
        done++;

        sleepms(waitms);
    }

    close(sd);
    return(0);
}



int fgetz(FILE *fd, u8 *data, int size) {
    u8     *p;

    if(!fgets(data, size, fd)) return(-1);
    for(p = data; *p && (*p != '\n') && (*p != '\r'); p++);
    *p = 0;
    return(p - data);
}



u8 *get_next(FILE *fd, in_addr_t *hosts_ip) {
    int     c;
    u8      *p,
            *l;
    static  u8  buff[1024];

    if(tabmode) {
        p = (void *)buff;
        for(;;) {
            c = fgetc(fd);
            if(c < 0) {
                if(p == buff) return(NULL);
                break;
            }
            if((c == '\t') || (c == ' ') || (c == '\r') || (c == '\n') || (c == ',') || (c == '|')) {
                if(p == buff) continue;
                if(tabmode == 1) tabmode = c;   // guessing the delimiter is the only thing to do
                break;
            }
            if((p - buff) >= sizeof(buff)) break;
            *p++ = c;
        }
        *p = 0;
        p = (void *)buff;
        l = strstr(p, "://");
        if(l) p = l + 3;
        l = strchr(p, '/');
        if(l) *l = 0;
        return(p);
    }

    while(fgetz(fd, buff, sizeof(buff)) >= 0) {
        if(!buff[0]) return("");
        if(strchr(buff, '#')) return(buff);                 // preserve comments
        for(p = (void *)buff; *p && ((*p == '\t') || (*p == ' ')); p++);    // skip spaces at the beginning
        if(!*p || (*p == '#') || (*p == ';') || (*p == '/')) continue;
        if(hosts_ip) *hosts_ip = inet_addr(p);
        for(; *p && (*p != '\t') && (*p != ' '); p++);      // skip 127.0.0.1
        if(!*p) continue;
        for(; *p && ((*p == '\t') || (*p == ' ')); p++);    // skip spaces again
        if(!*p || (*p == '#')) continue;
        l = strstr(p, "://");
        if(l) p = l + 3;
        l = strchr(p, '/');
        if(l) *l = 0;
        return(p);
    }
    return(NULL);
}



void put_list(FILE *fd, list_t *list) {
    int     i;

    for(i = 0; list[i].host; i++) {
        if(list[i].comment) {     // preserve comments
            fprintf(fd, "%s\n", list[i].host);
            continue;
        }
        if(!list[i].ok) {
            if(fddel) {
                if(tabmode) {
                    fprintf(fddel, "%s%c", list[i].host, tabmode);
                } else {
                    fprintf(fddel, "%s\t%s\n", inet_ntoa(*(struct in_addr *)&list[i].hosts_ip), list[i].host);
                }
            }
            continue;
        }
        if(tabmode) {
            fprintf(fd, "%s%c", list[i].host, tabmode);
        } else {
            fprintf(fd, "%s\t%s\n", inet_ntoa(*(struct in_addr *)&list[i].hosts_ip), list[i].host);
        }
    }
}



void *get_list(FILE *fd, int tot) {
    list_t  *list;
    int     i;
    u8      *p;

    if(tot < 0) {
        for(tot = 0;; tot++) {
            p = get_next(fd, NULL);
            if(!p) break;
        }
        return((void *)tot);
    }

    list = malloc(sizeof(list_t) * (tot + 1));
    if(!list) std_err();

    for(i = 0; i < tot; i++) {
        list[i].host = strdup(get_next(fd, &list[i].hosts_ip));
        if(!list[i].host) break;
        if(!list[i].host[0] || strchr(list[i].host, '#')) {
            list[i].comment = 1;
        } else {
            list[i].comment = 0;
        }
        list[i].ok = 1;
    }
    list[i].host = NULL;
    return(list);
}



int putns(u8 *data, u8 *str, int max) {
    int     len;
    u8      *i,
            *o,
            *p;

    i = str;
    o = data;
    for(p = str; *p; i = p + 1) {
        p = strchr(i, '.');
        if(!p) p = i + strlen(i);
        len = p - i;
        if((o + 1 - data) >= max) {
            fprintf(fdlog, "\nError: host \"%s\" is too long\n", str);
            exit(1);
        }
        *o++ = len;
        memcpy(o, i, len);
        o += len;
    }
    *o++ = 0;
    return(o - data);
}



int getxx(u8 *data, u32 *ret, int bits) {
    u32     num;
    int     i,
            bytes;

    bytes = bits >> 3;
    for(num = i = 0; i < bytes; i++) {
        num |= (data[i] << ((bytes - i - 1) << 3));
    }
    *ret = num;
    return(bytes);
}



int putxx(u8 *data, u32 num, int bits) {
    int     i,
            bytes;

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        data[i] = num >> ((bytes - i - 1) << 3);
    }
    return(bytes);
}



int timeout(int sock, int secs) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec  = secs;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



    // RFC 1035, note that this implementation is very easy
u32 resolv(int sd, int tid, struct sockaddr_in *peer, int query, char *host) {
    struct sockaddr_in peerl;
    in_addr_t   ip;
    u32     num,
            qdc,
            anc,
            rcode,
            typ;
    int     i,
            mysd,
            slen,
            rlen,
            err,
            psz;
    u8      sbuff[1024],
            rbuff[1024],
            *limit,
            *p,
            *errstr;

    p = sbuff;
    p += putxx(p, tid,      16);    // transaction ID
    p += putxx(p, 0x0100,   16);    // flags
    p += putxx(p, 1,        16);    // questions
    p += putxx(p, 0,        16);    // answer RRs
    p += putxx(p, 0,        16);    // authority RRs
    p += putxx(p, 0,        16);    // additional RRs
    p += putns(p, host,     sizeof(sbuff) - (p - sbuff));
    p += putxx(p, query,    16);
    p += putxx(p, C_IN,     16);
    slen = p - sbuff;

    mysd = 0;
    if(!sd) {
        mysd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if(mysd < 0) std_err();
        sd = mysd;
    }

    err = 0;
redo:
    for(i = 0; ; i++) {
        if(sendto(sd, sbuff, slen, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in)) >= 0) break;
        if(i >= 2) std_err();
        sleep(ONESEC);
    }
    for(;;) {
        if(timeout(sd, 3) < 0) {
            if(err >= 2) {
                fprintf(fdlog, "\nAlert: socket timeout, no reply from the DNS server for host %s\n", host);
                if(mysd) exit(1);
                return(htonl(INADDR_NONE));
            }
            err++;
            sleep(ONESEC);
            goto redo;
        }
        psz = sizeof(struct sockaddr_in);
        rlen = recvfrom(sd, rbuff, sizeof(rbuff), 0, (struct sockaddr *)&peerl, &psz);
        if(rlen < 0) std_err(); // not a good solution because a big reply kills the tool
        if(peerl.sin_addr.s_addr != peer->sin_addr.s_addr) continue;

        p = rbuff;
        limit = rbuff + rlen;

        p += getxx(p, &num, 16);        if(p >= limit) continue;    // transaction ID
        if(num != tid) continue;
        p += getxx(p, &num, 16);        if(p >= limit) continue;    // flags
        if(!(num & 0x8000)) continue;
        rcode = num & 0x000f;
        switch(rcode) {
            case 0: {              errstr = NULL; break; }
            case 1: { dns_error++; errstr = "Format error (1) - The name server was unable to interpret the query."; break; }
            case 2: {              errstr = "Server failure (2) - The name server was unable to process this query due to a problem with the name server."; break; }
            case 3: {              errstr = ": hostname can't be resolved"; break; /* errstr = "Name Error - Meaningful only for responses from an authoritative name server, this code signifies that the domain name referenced in the query does not exist."; break; */ }
            case 4: { dns_error++; errstr = "Not Implemented (4) - The name server does not support the requested kind of query."; break; }
            case 5: { dns_error++; errstr = "Refused (5) - The name server refuses to perform the specified operation for policy reasons.  For example, a name server may not wish to provide the information to the particular requester, or a name server may not wish to perform a particular operation (e.g., zone transfer) for particular data."; break; }
            default:{ dns_error++; errstr = "unknown error"; break; }
        }
        if(errstr) {
            if(verbose) fprintf(fdlog, "\nHost %s: DNS %s\n", host, errstr);
            return(htonl(INADDR_NONE));
            //exit(1);
        }
        if(query != T_A) return(1);                                 // any other reply is ok for non A records

        p += getxx(p, &qdc, 16);        if(p >= limit) continue;    // questions
        p += getxx(p, &anc, 16);        if(p >= limit) continue;    // answer RRs
        p += 2 + 2;                     if(p >= limit) continue;    // authority RRs + additional RRs
        for(i = 0; i < qdc; i++) {
            if(*p & 0xc0) {             // easy mode
                p += 2;
            } else {
                while(*p) p++;
            }                           if(p >= limit) continue;    // name
            p += 2 + 2;                 if(p >= limit) continue;    // type + class
        }
        for(i = 0; i < anc; i++) {
            if(*p & 0xc0) {             // easy mode
                p += 2;
            } else {
                while(*p) p++;
            }                           if(p >= limit) continue;    // name
            p += getxx(p, &typ, 16);    if(p >= limit) continue;    // type
            p += 2 + 4;                 if(p >= limit) continue;    // class + TTL
            p += getxx(p, &num, 16);    if(p >= limit) continue;    // data length
            if((query == T_A) && (num == 4) && (typ == query)) {
                getxx(p, &ip,   32);
                return(ip);
            }
            p += num;                   if(p >= limit) continue;
        }
    }

    if(mysd) close(mysd);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


