/*
    Copyright 2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#define GTK_ENABLE_BROKEN
#include <gtk/gtk.h>



#ifdef WIN32
    #include <direct.h>
    #include <windows.h>

    #define WINWIDTH        654
    #define WAITHELPDATA    500
    HANDLE  process_pid = 0;
    int     detach = 0;
    HANDLE  out_read,
            in_write;
#else
    #include <unistd.h>
    #include <sys/types.h>
    #include <fcntl.h>

    #define stricmp strcasecmp
    #define stristr strcasestr
    #define WINWIDTH        580
    #define WAITHELPDATA    500000
    #ifndef WNOHANG
        #define WNOHANG     1
        #define SIGTERM     15
        #define SIGKILL     9
    #endif
    int     process_pid = 0,
            stdin_pipe[2],
            stdout_pipe[2],
            stderr_pipe[2],
            stdout_close,
            stderr_close;
    char    *ptr_exec[128];
#endif



#define VER             "thegui v.0.3"
#define BUFFSZ          16384
#define FNAMESZ         512
#define ARGSZ           FNAMESZ
#define FONTSZ          12
#define REFRESH         100
#define RUNLABEL        "    R U N    "
#define NOTRUNLABEL     "...running..."
#define STRCPY(x,y)     mystrcpy(x, y, sizeof(x));
#define SKIPOPTIONS     if(!stricmp(args[i], "[option]") || !stricmp(args[i], "[options]")) continue



void startwindow(void);



GtkWidget   *window,
            **arg_entry,
            **opt_entry,
            *manual_entry,
            *view,
            *runlabel,
            *inputentry;
GdkFont     *fixed_font;
int         argsnum,
            optsnum,
            opts_row,
            usage_off,
            noargs,
            newstyle,
            viewsize;
u_char      **args,
            *exe,
            *filename,
            help[BUFFSZ],
            cmdbuff[BUFFSZ],
            exename[FNAMESZ];
struct options {
    u_char  *cmd;
    int     type;
    u_char  val[ARGSZ];
    u_char  *desc;
    int     on;
} *opts;



int mystrcpy(u_char *dst, u_char *src, int max) {
    u_char  *p = dst;

    while(*src && --max) {
        *p++ = *src++;
    }
    *p = 0;
    return(p - dst);
}



void dialogbox(char *text, char *head) {
    GtkWidget   *dlgwin,
                *label;

    dlgwin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_position(GTK_WINDOW(dlgwin), GTK_WIN_POS_CENTER);
    gtk_window_set_title(GTK_WINDOW(dlgwin), head);
    gtk_container_set_border_width(GTK_CONTAINER(dlgwin), 10);
    gtk_window_set_policy(GTK_WINDOW(dlgwin), FALSE, FALSE, FALSE);
    gtk_signal_connect(GTK_OBJECT(dlgwin), "destroy", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
    gtk_signal_connect(GTK_OBJECT(dlgwin), "delete_event", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dlgwin));

    label = gtk_label_new(text);
    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
    gtk_container_add(GTK_CONTAINER(dlgwin), label);

    gtk_widget_show_all(dlgwin);
    gtk_main();
}



void std_err(void) {
    dialogbox(strerror(errno), VER);
    exit(1);
}



#ifdef WIN32
void winerr(void) {
    char    message[512];

    if(FormatMessage(
      FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      GetLastError(),
      0,
      (char *)&message,
      sizeof(message),
      NULL)) {
        dialogbox(message, VER);
        exit(1);
    }

    std_err();
}
#endif



void file_ok_sel(GtkWidget *widget, gpointer data) {
    mystrcpy(filename, (u_char *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(data)), FNAMESZ);
    gtk_widget_destroy(GTK_WIDGET(data));
    gtk_main_quit();
}



void select_file(u_char *desc) {
    GtkWidget   *filew;

    filew = gtk_file_selection_new(desc);
    gtk_window_set_position(GTK_WINDOW(filew), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(filew), "destroy", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "clicked", (GtkSignalFunc)file_ok_sel, filew);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(filew));
    gtk_widget_show_all(filew);
    gtk_main();
}



void get_file(GtkWidget *widget, gpointer data) {
    u_char  saveme[FNAMESZ];

    *saveme  = 0;
    filename = saveme;
    select_file("select the file from the window at right");
    if(!*saveme) return;
    gtk_entry_set_text(GTK_ENTRY(data), saveme);
}



void get_folder(GtkWidget *widget, gpointer data) {
    u_char  saveme[FNAMESZ];

    *saveme  = 0;
    filename = saveme;
    select_file("select the folder from the left window with double-click");
    if(!*saveme) return;
    gtk_entry_set_text(GTK_ENTRY(data), saveme);
}



void save_output(void) {
    FILE    *fd;
    u_int   *ldata;
    int     len;
    u_char  saveme[FNAMESZ],
            *data;      // WORK-AROUND: on Win gtk_get_editable_chars
                        // fails and text.ch contains u_int chars!!!

    *saveme = 0;
    filename = saveme;
    select_file("save the output in a file");
    if(!*saveme) return;
    fd = fopen(saveme, "wb");
    if(!fd) std_err();
    data  = (u_char *)GTK_TEXT(view)->text.ch;
    ldata = (u_int *)data;
    len   = gtk_text_get_length(GTK_TEXT(view));

    if(*ldata > 0xff) {
        fputs(data, fd);
    } else {
        while(len--) {
            fputc(*ldata & 0xff, fd);
            ldata++;
        }
    }

    fclose(fd);
}



void find_row_usage(void) {
    int     i,
            avg    = 0,
            rem    = 0;
    u_char  *p,
            *l,
            *d,
            *u,
            *input,
            posnum = 0,
            pos[16],
            pos2[256];

    input = strdup(help);

    usage_off = 0;

    memset(pos, 0, sizeof(pos));

    p = input;
    do {
        l = strchr(p, '\n');
        if(l) *l = 0;

        if(!usage_off) {
            u = stristr(p, "usage");
            if(u) {
                d = strchr(u, ':');
                if(d) {
                    usage_off = d + 1 - input;
                } else {
                    usage_off = u + 5 - input;
                }
            }
        }

        d = strchr(p, '-');
        if(d) {
            while(*d && (*d > ' ')) d++;    // pass the option
            while(*d && (*d <= ' ')) d++;   // pass the space after the option
            if(strstr(d, "  ")) {
                while(strstr(d, "  ")) d++;
                d++;
            } else {
                if(*d == '[') {
                    u = strchr(d + 1, ']');
                    if(u) d = u + 1;
                } else if(*d == '<') {
                    u = strchr(d + 1, '>');
                    if(u) d = u + 1;
                }
            }
            pos[posnum] = d - p;
            if(++posnum == sizeof(pos)) break;
        }

        p = l + 1;
    } while(l);

    memset(pos2, 0, sizeof(pos2));

    for(i = 0; i < posnum; i++) {
        if(pos[i] > 40) continue;   // useless, 40 is half window
        pos2[pos[i]]++;
    }

    for(i = 0; i < sizeof(pos2); i++) {
        if(pos2[i] > pos2[avg]) avg = i;
    }

    for(i = 0; i < sizeof(pos2); i++) {
        if(pos2[i] && (pos2[i] != pos2[avg])) rem++;
    }

    if(rem >= (posnum / 3)) avg = 0;
    if((posnum < 2) || stristr(help, "invalid option")) avg = 0;
    opts_row = avg;

    free(input);
}



int get_args(void) {
    int     num  = 0,
            stop = 0;
    u_char  *p,
            *l,
            *input;

    if(!usage_off) return(0);

    input = strdup(help);

    p = input + usage_off;
    l = strchr(p, '\n');
    if(l) {
        memset(l, 0, 5);    // required!!!
    }
    l = p;

    do {
        for(p = l + 1; *p && (*p <= ' '); p++);
        if(!*p) break;

        if(*p == '[') {
            l = strchr(p + 1, ']');
            if(l) {
                l++;
                *l = 0;
            } else {
                stop = 1;
            }
        } else if(*p == '<') {
            l = strchr(p + 1, '>');
            if(l) {
                l++;
                *l = 0;
            } else {
                stop = 1;
            }
        } else {
            for(l = p + 1; *l && (*l > ' '); l++);
            if(*l) {
                *l = 0;
            } else {
                stop = 1;
            }
        }

        args = realloc(args, sizeof(u_char *) * (num + 1));
        if(!args) std_err();
        args[num] = malloc(ARGSZ);
        if(!args[num]) std_err();
        mystrcpy(args[num], p, ARGSZ);
        num++;
    } while(!stop);

    free(input);
    return(num);
}



int get_opts(void) {
    int     num = 0,
            len1,
            len2;
    u_char  *p,
            *l,
            *t,
            *t2,
            *start_line,
            *input;

    input = strdup(help);

    p = strchr(input + usage_off, '\n');
    if(!p) goto quit;
    p++;

    opts = NULL;

    while((l = strchr(p, '\n'))) {
        *l = 0;
        start_line = p;
        if(*p == '*') p++;
        while(*p && (*p <= ' ')) p++;
        if(*p == '-') {
            for(t = p + 1; *t && (*t > ' '); t++);
            *t = 0;

            if(*(t - 1) == ':') *(t - 1) = 0;

            t2 = strchr(p, ',');
            if(t2) {
                p = t2 + 1;
                if(*p <= ' ') p++;
                for(t = p + 1; *t && (*t > ' '); t++);
                *t = 0;
            }

            opts = realloc(opts, sizeof(struct options) * (num + 1));
            if(!opts) std_err();
            opts[num].cmd = strdup(p);

            if(t[1] <= ' ') {
                opts[num].type = 0;
            } else {
                opts[num].type = 1;

                p = t + 1;
                if(opts_row) {
                    t = start_line + opts_row - 1;
                    while(*t && (*t <= ' ')) t--;
                    t++;
                } else {
                    for(t = p; *t && (*t > ' '); t++);
                }
                *t = 0;

                if(*(t - 1) == ':') *(t - 1) = 0;

                if((!opts_row) && (*p != '<') && (*p != '[')) {
                    opts[num].type = 0;
                    opts[num].val[0] = 0;
                    *t = ' ';
                    t = p - 1;
                    goto skip;
                }

                STRCPY(opts[num].val, p);
            }
skip:
            for(p = t + 1; *p && (*p <= ' '); p++);

            opts[num].desc = strdup(p);

            opts[num].on = 0;

            num++;
        } else {
            if(opts_row) {
                for(t = start_line + opts_row; *p && (*p <= ' '); p++);
                if(p == t) {
                    len1 = strlen(opts[num - 1].desc);
                    len2 = strlen(p);
                    opts[num - 1].desc = realloc(opts[num - 1].desc, len1 + 1 + len2 + 1);
                    sprintf(opts[num - 1].desc + len1, "\n%s", p);
                }
            }
        }
        p = l + 1;
    }

quit:
    free(input);
    return(num);
}



int format_text(u_char *data, int len) {
    int     pos = 0,
            tab;
    u_char  *start = data;

    while(len--) {
        if(*data == '\t') {
            tab = 8 - (pos & 7);
            memmove(data + tab, data + 1, len);
            memset(data, ' ', tab);
            data += tab;
            pos  += tab;
            continue;

        } else if(*data == '\n') {
            pos = -1;

        } else {
            if((*data < ' ') || (*data >= 0x7f)) *data = ' ';
        }

        data++;
        pos++;
    }

    return(data - start);
}



void stop_process(GtkWidget *widget, gpointer data) {
    if(!process_pid) return;

#ifdef WIN32
    TerminateProcess(process_pid, -1);
#else
    waitpid(process_pid, NULL, WNOHANG);
    kill(process_pid, SIGTERM);
//    kill(process_pid, SIGKILL);
#endif

    process_pid = 0;
    gtk_label_set_text(GTK_LABEL(runlabel), RUNLABEL);
    gtk_window_set_title(GTK_WINDOW(window), exe);
}



void write_output(u_char *buff, int len) {
    gtk_text_insert(GTK_TEXT(view), fixed_font, NULL, NULL, buff, len);
    gtk_adjustment_set_value(
        GTK_TEXT(view)->vadj,
        GTK_TEXT(view)->vadj->upper - GTK_TEXT(view)->vadj->page_size);
}



void cmdbind(u_char *cmd, u_char *out, int maxout) {
    int     bufflen;
    u_char  buff[BUFFSZ];

#define PREOP   \
    maxout >>= 1;                   /* for \t handling */   \
    bufflen = sizeof(buff) >> 1;    /* not sanitized!  */   \
    if(process_pid) goto readcmd; \
    gtk_label_set_text(GTK_LABEL(runlabel), NOTRUNLABEL);

#ifdef WIN32
    SECURITY_ATTRIBUTES sec;
    STARTUPINFO         startup_info;
    PROCESS_INFORMATION process_info;
    HANDLE  in_read,
            out_write;
    DWORD   pstat,
            len;

    PREOP

    sec.nLength              = sizeof(sec);
    sec.lpSecurityDescriptor = NULL;
    sec.bInheritHandle       = TRUE;
    if(!CreatePipe(&out_read, &out_write, &sec, 0)) winerr();
    if(!CreatePipe(&in_read,  &in_write,  &sec, 0)) winerr();

    GetStartupInfo(&startup_info);
    startup_info.dwFlags     = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
    startup_info.wShowWindow = SW_HIDE;
    startup_info.hStdInput   = in_read;
    startup_info.hStdOutput  = out_write;
    startup_info.hStdError   = out_write;
 
    if(!CreateProcess(
        NULL,                               // lpApplicationName
        cmd,                                // lpCommandLine
        &sec,                               // lpProcessAttributes
        &sec,                               // lpThreadAttributes
        TRUE,                               // bInheritHandles
        detach | CREATE_NEW_PROCESS_GROUP,  // dwCreationFlags
        NULL,                               // lpEnvironment
        NULL,                               // lpCurrentDirectory
        &startup_info,                      // lpStartupInfo
        &process_info)) {                   // lpProcessInformation
            winerr();
    }
    process_pid = process_info.hProcess;

    if(!CloseHandle(out_write)) winerr();
    if(!CloseHandle(in_read)) winerr();

    if(out) {
        sleep(WAITHELPDATA);    // useful

        if(PeekNamedPipe(out_read, NULL, 0, NULL, (DWORD *)&len, 0) && len) {
            ReadFile(out_read, out, maxout, (DWORD *)&len, NULL);
            len = format_text(out, len);
            out[len] = 0;
        }
        stop_process(NULL, NULL);
        return;
    }

readcmd:
    if(PeekNamedPipe(out_read, NULL, 0, NULL, (DWORD *)&len, 0) && len) {
        ReadFile(out_read, buff, bufflen, (DWORD *)&len, NULL);
        if(len <= 0) {
            stop_process(NULL, NULL);
            return;
        }
        write_output(buff, format_text(buff, len));

    } else if(!GetExitCodeProcess(process_pid, &pstat) || (pstat != STILL_ACTIVE)) {
        stop_process(NULL, NULL);
        return;
    }

#else

    int     len;

    PREOP

    if(pipe(stdin_pipe) || pipe(stdout_pipe) || pipe(stderr_pipe)) {
        std_err();
    }

    process_pid = fork();
    if(process_pid < 0) {
        std_err();

    } else if(!process_pid) {
        close(0);
        dup(stdin_pipe[0]);
        close(stdin_pipe[0]);
        close(stdin_pipe[1]);

        close(1);
        dup(stdout_pipe[1]);
        close(stdout_pipe[0]);
        close(stdout_pipe[1]);

        close(2);
        dup(stderr_pipe[1]);
        close(stderr_pipe[0]);
        close(stderr_pipe[1]);

        if(execvp(ptr_exec[0], ptr_exec) < 0) {
            kill(getppid(), SIGTERM);   // std_err()
        }
        exit(EXIT_FAILURE);
    }

    close(stdin_pipe[0]);
    close(stdout_pipe[1]);
    close(stderr_pipe[1]);

    if(fcntl(stdout_pipe[0], F_SETFL, fcntl(stdout_pipe[0], F_GETFL) | O_NONBLOCK) < 0) std_err();
    if(fcntl(stderr_pipe[0], F_SETFL, fcntl(stderr_pipe[0], F_GETFL) | O_NONBLOCK) < 0) std_err();

    if(out) {
        usleep(WAITHELPDATA);   // useful

        len = read(stderr_pipe[0], out, maxout);    // STDERR
        if(len > 0) {
            len = format_text(out, len);
            out += len;
        }
        len = read(stdout_pipe[0], out, maxout);    // STDOUT
        if(len > 0) {
            len = format_text(out, len);
            out += len;
        }
        *out = 0;
        stop_process(NULL, NULL);
        return;
    }

    stdout_close = 0;
    stderr_close = 0;

readcmd:
    len = read(stdout_pipe[0], buff, bufflen);      // STDOUT
    if(len > 0) {
        write_output(buff, format_text(buff, len));
    } else if(!len) {
        stdout_close = 1;
    }

    len = read(stderr_pipe[0], buff, bufflen);      // STDOUT
    if(len > 0) {
        write_output(buff, format_text(buff, len));
    } else if(!len) {
        stderr_close = 1;
    }

    if(stdout_close && stderr_close) {              // stop if finished
        stop_process(NULL, NULL);
        return;
    }

#endif
#undef PREOP
}



void activate(GtkWidget *widget, gpointer data) {
    opts[(int)data].on = !opts[(int)data].on;
}



void clear_view(GtkWidget *widget, gpointer data) {
    gtk_text_set_point(GTK_TEXT(view), 0);
    gtk_text_forward_delete(GTK_TEXT(view), gtk_text_get_length(GTK_TEXT(view)));
}



void show_help(GtkWidget *widget, gpointer data) {
    dialogbox(help, exe);
}



void show_about(GtkWidget *widget, gpointer data) {
    dialogbox(
        VER "\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    http://aluigi.altervista.org\n"
        " \n"
        "This program has 3 ways for being used:\n"
        " \n"
        "1  when launched select the program's executable you want to load\n"
        " \n"
        "2  drag'n'drop. Press the left mouse button over the executable of the program to\n"
        "   load and move it over the executable of \""VER"\" while keeping the button\n"
        "   pressed\n"
        " \n"
        "3  move the executable of \""VER"\" in the Sendto folder of your system and then\n"
        "   press the right mouse button over the program's executable to load and select\n"
        "   this tool from the SendTo menu\n"
        " \n"
        "Note that for obvious reasons this program cannot work with all the existent\n"
        "command-line programs but in my tests it was able to handle many of the tools that\n"
        "show the list of available options with the '-' char before them (like all my\n"
        "tools for example or netcat and nmap and some others)\n"
        "\n",
        VER);
}



gint runcmd(gpointer data) {
    if(!process_pid) return(TRUE);
    cmdbind(cmdbuff, NULL, 0);
    return(TRUE);
}



int getthefirst(void) {
    if(argsnum > 1) {
        if((args[1][0] != '[') && (args[1][0] != '<')) return(0);
        return(1);
    }
    return(0);
}



void run_process(GtkWidget *widget, GtkWidget *entry) {
    int     i,
            len;
    u_char  *data;
#ifndef WIN32
    int     num_exec;
    u_char  viscmdbuff[BUFFSZ],
            *p,
            *l,
            *limit;
#endif

    if(process_pid) return;

#ifndef WIN32
    num_exec = 0;
    ptr_exec[num_exec++] = cmdbuff;
#endif
    len = sprintf(cmdbuff, "%s ", exe);

    data = (u_char *)gtk_entry_get_text(GTK_ENTRY(manual_entry));
#ifdef WIN32
    len += sprintf(
        cmdbuff + len,
        "%s ",
        data);
#else
    limit = data + strlen(data);
    while(data < limit) {
        if(*data == '\"') {
            data++;
            p = strchr(data, '\"');
            if(!p) p = limit;
        } else {
            p = strchr(data, ' ');
            if(!p) p = strchr(data, '\t');
            if(!p) p = limit;
        }
        if(p == data) {
            data++;
            continue;
        }

        cmdbuff[len - 1] = 0;
        ptr_exec[num_exec++] = cmdbuff + len;
        len += sprintf(
            cmdbuff + len,
            "%.*s ",
            p - data, data);

        data = p + 1;
    }
#endif

    for(i = 0; i < optsnum; i++) {
        if(!opts[i].on) continue;

#ifdef WIN32
        len += sprintf(
            cmdbuff + len,
            "%s ",
            opts[i].cmd);

        if(opts[i].type) {
            data = (u_char *)gtk_entry_get_text(GTK_ENTRY(opt_entry[i]));
            len += sprintf(
                cmdbuff + len,
                "%s%s%s ",
                (*data == '\"') ? "" : "\"",
                data,
                (*data == '\"') ? "" : "\"");
        }
#else
        cmdbuff[len - 1] = 0;
        ptr_exec[num_exec++] = cmdbuff + len;
        len += sprintf(
            cmdbuff + len,
            "%s ",
            opts[i].cmd);

        if(opts[i].type) {
            cmdbuff[len - 1] = 0;
            ptr_exec[num_exec++] = cmdbuff + len;
            data = (u_char *)gtk_entry_get_text(GTK_ENTRY(opt_entry[i]));
            len += sprintf(
                cmdbuff + len,
                "%s ",
                data);
        }
#endif
    }

    for(i = getthefirst(); i < argsnum; i++) {
        SKIPOPTIONS;

        data = (u_char *)gtk_entry_get_text(GTK_ENTRY(arg_entry[i]));
        if(!*data) continue;

#ifdef WIN32
        len += sprintf(
            cmdbuff + len,
            "%s%s%s ",
            ((*data == '\"') || noargs) ? "" : "\"",
            data,
            ((*data == '\"') || noargs) ? "" : "\"");
#else
        cmdbuff[len - 1] = 0;
        ptr_exec[num_exec++] = cmdbuff + len;
        len += sprintf(
            cmdbuff + len,
            "%s ",
            data);
#endif
    }

#ifndef WIN32
    cmdbuff[len - 1] = 0;
    ptr_exec[num_exec] = NULL;
    memcpy(viscmdbuff, cmdbuff, len);
    l = viscmdbuff + len - 1;
    for(p = viscmdbuff; p < l; p++) {
        if(!*p) *p = ' ';
    }
    gtk_window_set_title(GTK_WINDOW(window), viscmdbuff);
#else
    gtk_window_set_title(GTK_WINDOW(window), cmdbuff);
#endif

    clear_view(NULL, NULL);
    cmdbind(cmdbuff, NULL, 0);
}



void myquit(GtkWidget *widget, gpointer data) {
    stop_process(NULL, NULL);
    gtk_main_quit();
}



void myinputfile(GtkWidget *widget, gpointer data) {
#ifdef WIN32
    DWORD   len;
#else
    int     len;
#endif
    FILE    *fd;
    u_char  buff[BUFFSZ],
            saveme[FNAMESZ];

    if(!process_pid) return;

    *saveme  = 0;
    filename = saveme;
    select_file("select the file to send to stdin");
    if(!*saveme) return;

    fd = fopen(saveme, "rb");
    if(!fd) return;

    while((len = fread(buff, 1, sizeof(buff), fd))) {
#ifdef WIN32
        WriteFile(in_write, buff, len, &len, NULL);
#else
        write(stdin_pipe[1], buff, len);
#endif
    }

    fclose(fd);
}



void myinput(GtkWidget *widget, gpointer data) {
#ifdef WIN32
    DWORD len;
#endif
    u_char  *input;

    if(!process_pid) return;

    input = (u_char *)gtk_entry_get_text(GTK_ENTRY(inputentry));
#ifdef WIN32
    WriteFile(in_write, input, strlen(input), &len, NULL);
    WriteFile(in_write, "\r\n", 2, &len, NULL);
#else
    write(stdin_pipe[1], input, strlen(input));
    write(stdin_pipe[1], "\n", 1);
#endif

    gtk_entry_set_text(GTK_ENTRY(inputentry), "");
}



int load_help(void) {
    int     i,
            bigger,
            helpssz[10];    // same or bigger amount of helps
    u_char  cmd[FNAMESZ + 20],
            *helps[] = {
                "",
                "--help",
                "-h",
                "-?",
                "/?",
                "/h",
                "/help",
                NULL
            };

    for(i = 0; helps[i]; i++) {
#ifdef WIN32
        sprintf(cmd, "%s %s", exe, helps[i]);
#else
        ptr_exec[0] = exe;
        if(helps[i][0]) {   // work-around
            ptr_exec[1] = helps[i];
        } else {
            ptr_exec[1] = NULL;
        }
        ptr_exec[2] = NULL;
#endif
        cmdbind(cmd, help, sizeof(help));
        helpssz[i] = strlen(help);
        find_row_usage();
        if(usage_off || opts_row) break;
    }

    if(!usage_off || !opts_row) {
        for(bigger = i--; i >= 0; i--) {
            if(helpssz[i] > helpssz[bigger]) bigger = i;
        }
#ifdef WIN32
        sprintf(cmd, "%s %s", exe, helps[bigger]);
#else
        ptr_exec[0] = exe;
        if(helps[bigger] && helps[bigger][0]) {
            ptr_exec[1] = helps[bigger];
        } else {
            ptr_exec[1] = NULL;
        }
        ptr_exec[2] = NULL;
#endif
        cmdbind(cmd, help, sizeof(help));
        find_row_usage();
    }

    return(0);
}



int calcoutputsize(int size) {
    int     i;
    u_char  *p;

#define CALCSZ  size -= (FONTSZ + 3)

    CALCSZ;

    for(i = getthefirst(); i < argsnum; i++) {
        CALCSZ;
    }

    for(i = 0; i < optsnum; i++) {
        for(p = opts[i].desc; *p; p++) {
            if(*p == '\n') {
                CALCSZ;
            }
        }
        CALCSZ;
    }

    return(size);
#undef CALCSZ
}



void redostyle(GtkWidget *widget, gpointer data) {
    newstyle = !newstyle;
    gtk_widget_destroy(window);
    startwindow();
}



void free_all(void) {
    int     i;

    for(i = 0; i < argsnum; i++) {
        free(args[i]);
    }
    free(args);

    for(i = 0; i < optsnum; i++) {
        free(opts[i].cmd);
        free(opts[i].desc);
    }
    free(opts);

    free(arg_entry);
    free(opt_entry);
}



void loadfile(void) {
    u_char      *dir,
                *p;

    dir = exename;
    exe = exename;
    p = strrchr(dir, '\\');
    if(!p) p = strrchr(dir, '/');
    if(p) {
        *p = 0;
        exe = p + 1;
    }
    if(dir != exe) {
        if(chdir(dir) < 0) std_err();
        if(*(exe - 2) == ':') chdir("\\");
    }

    load_help();

    noargs  = 0;
    argsnum = get_args();
    optsnum = get_opts();

    if(!argsnum) noargs = 1;

    arg_entry = malloc(argsnum * sizeof(GtkWidget *));
    opt_entry = malloc(optsnum * sizeof(GtkWidget *));

    viewsize = 400 + ((gdk_screen_width() - 640) / 2);
    if(viewsize < 0) viewsize = 400;
    viewsize = calcoutputsize(viewsize);
    if(viewsize < 200) {
        newstyle = 1;
        viewsize = 600;
    } else {
        newstyle = 0;
    }
}



void reloadfile(GtkWidget *widget, gpointer data) {
    free_all();

    *exename = 0;
    filename = exename;
    select_file("select or type the name of the command-line program to load");
    if(!*exename) return;
    loadfile();

    gtk_widget_destroy(window);
    startwindow();
}




void startwindow(void) {
#define HBOX            hbox = gtk_hbox_new(FALSE, 0);                                                          \
                        gtk_container_add(GTK_CONTAINER(vbox), hbox);
#define SEP             separator = gtk_hseparator_new();                                                       \
                        gtk_box_pack_start(GTK_BOX(vbox), separator, FALSE, FALSE, 0);
#define TOOLTIP(x,y)    tooltip = gtk_tooltips_new();                                                           \
                        gtk_tooltips_set_delay(tooltip, 1500);                                                  \
                        gtk_tooltips_set_tip(tooltip, x, y, NULL);
#define BUTTON(x,y,z)   button = gtk_button_new_with_label(x);                                                  \
                        TOOLTIP(button, z)                                                                      \
                        gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(y), NULL);            \
                        gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
#define BUTTON2(x,y,z)  button = gtk_button_new_with_label(x);                                                  \
                        TOOLTIP(button, z)                                                                      \
                        gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(y), NULL);            \
                        gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
#define LABEL(x)        label = gtk_label_new(x);                                                               \
                        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);
#define SET_PASS(x,y)   if(stristr(x, "password")) {                                                            \
                            gtk_entry_set_visibility(GTK_ENTRY(y), FALSE);                                      \
                        }
#define SET_FILE(x,y)   if(stristr(x, "directory") || stristr(x, "folder")) {                                   \
                            button = gtk_button_new_with_label("select");                                       \
                            gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(get_folder), y);  \
                            gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);                         \
                        } else if(stristr(x, "file")) {                                                         \
                            button = gtk_button_new_with_label("select");                                       \
                            gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(get_file), y);    \
                            gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);                         \
                        }

    GtkWidget   *vbox,
                *hbox,
                *paned,
                *button,
                *label,
                *scroll,
                *separator;
    GtkTooltips *tooltip;
    int         i;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
    gtk_window_set_title(GTK_WINDOW(window), VER);
    gtk_container_set_border_width(GTK_CONTAINER(window), 0);
    gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);
    gtk_window_set_default_size(GTK_WINDOW(window), -1, -1);
    gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(myquit), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(window));

    if(newstyle) {
        paned = gtk_hpaned_new();
    } else {
        paned = gtk_vpaned_new();
    }
    gtk_container_add(GTK_CONTAINER(window), paned);

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_paned_add1(GTK_PANED(paned), vbox);
    TOOLTIP(paned, "keep the mouse button pressed for resizing the options window");

    HBOX
    manual_entry = gtk_entry_new();
    TOOLTIP(manual_entry, "type the other command-line parameters you need, they will be added to the other selected options");
    gtk_signal_connect(GTK_OBJECT(manual_entry), "activate", GTK_SIGNAL_FUNC(run_process), NULL);
    gtk_entry_set_max_length(GTK_ENTRY(manual_entry), ARGSZ);
    gtk_box_pack_start(GTK_BOX(hbox), manual_entry, TRUE, TRUE, 0);

    LABEL("add command-line")

    BUTTON2("stdin-file", myinputfile, "select a file to send to stdin")
    BUTTON2("stdin",      myinput,     "press to send the text at right to stdin (or simply press RETURN)")

    inputentry = gtk_entry_new();
    gtk_signal_connect(GTK_OBJECT(inputentry), "activate", GTK_SIGNAL_FUNC(myinput), NULL);
    gtk_box_pack_start(GTK_BOX(hbox), inputentry, TRUE, TRUE, 0);
    TOOLTIP(inputentry, "type your input text here and then press RETURN");

    for(i = getthefirst(); i < argsnum; i++) {
        SKIPOPTIONS;

        HBOX

        arg_entry[i] = gtk_entry_new();
        gtk_signal_connect(GTK_OBJECT(arg_entry[i]), "activate", GTK_SIGNAL_FUNC(run_process), NULL);
        gtk_entry_set_max_length(GTK_ENTRY(arg_entry[i]), ARGSZ);
#ifdef HIDEPASS
        SET_PASS(args[i], arg_entry[i])
#endif
        gtk_box_pack_start(GTK_BOX(hbox), arg_entry[i], TRUE, TRUE, 0);
        LABEL(args[i])
        SET_FILE(args[i], arg_entry[i])
    }

    if(optsnum) {
        SEP
    }

    for(i = 0; i < optsnum; i++) {
        HBOX

        opt_entry[i] = gtk_check_button_new_with_label(opts[i].desc);
        gtk_signal_connect(GTK_OBJECT(opt_entry[i]), "clicked", GTK_SIGNAL_FUNC(activate), (void *)i);
        gtk_box_pack_start(GTK_BOX(hbox), opt_entry[i], FALSE, FALSE, 0);

        if(opts[i].type) {
            opt_entry[i] = gtk_entry_new();
            gtk_signal_connect(GTK_OBJECT(opt_entry[i]), "activate", GTK_SIGNAL_FUNC(run_process), NULL);
#ifdef HIDEPASS
            SET_PASS(opts[i].desc, opt_entry[i])
#endif
            gtk_entry_set_text(GTK_ENTRY(opt_entry[i]), opts[i].val);
            gtk_box_pack_start(GTK_BOX(hbox), opt_entry[i], TRUE, TRUE, 0);
            SET_FILE(opts[i].desc, opt_entry[i])
        }
    }

    HBOX
    BUTTON("load",   reloadfile,   "load another executable")
    BUTTON("clear",  clear_view,   "clear the output screen")
    BUTTON("dump",   save_output,  "save the output screen in a file")

    BUTTON("stop",   stop_process, "stop the running program")

    runlabel = gtk_label_new(RUNLABEL);
    button = gtk_button_new();
    TOOLTIP(button, "start the program")
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(run_process), NULL);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(button), runlabel);

    BUTTON("help",   show_help,    "shows the help and options of the running program")

    BUTTON("visual", redostyle,    "change the visualization mode")
    BUTTON("about",  show_about,   "about THEGUI")
    BUTTON("quit",   myquit,       "quit THEGUI")

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_paned_add2(GTK_PANED(paned), hbox);
    view = gtk_text_new(NULL, NULL);
    gtk_widget_set_usize(view, WINWIDTH, viewsize);
    gtk_text_set_editable(GTK_TEXT(view),  FALSE);
    gtk_text_set_word_wrap(GTK_TEXT(view), TRUE);
    gtk_text_set_line_wrap(GTK_TEXT(view), TRUE);
    gtk_box_pack_start(GTK_BOX(hbox), view, TRUE, TRUE, 0);
    scroll = gtk_vscrollbar_new(GTK_TEXT(view)->vadj);
    gtk_box_pack_start(GTK_BOX(hbox), scroll, FALSE, FALSE, 0);

    gtk_widget_show_all(window);
    g_timeout_add(REFRESH, runcmd, NULL);
    gtk_main();
}



int main(int argc, char *argv[]) {
#ifdef WIN32
    struct  _OSVERSIONINFOA verinfo;
#endif

    setbuf(stdin,  NULL);
    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    gtk_init(&argc, &argv);

    fixed_font = gdk_font_load(
#ifdef WIN32
//        "-*-courier-medium-r-normal--12-120-75-75-m-70-iso8859-1");
        "-*-courier-medium-r-normal-*-*-10-*-*-m-*-*-*");
    verinfo.dwOSVersionInfoSize = sizeof(verinfo);      // useless but cool
    GetVersionEx(&verinfo);
    if(verinfo.dwPlatformId < VER_PLATFORM_WIN32_NT) {  // Win9x
        detach = DETACHED_PROCESS;
    }
#else
	// "-microsoft-verdana-medium-r-normal-*-*--15-140-*-*-p-93-koi8-uni");
	// "-misc-deja sans mono-medium-r-normal--15-140-75-75-m-90-koi8-uni");
	"-b&h-lucidatypewriter-medium-r-normal-sans-15-140-*-*-m-*-iso8859-1");
     //"-misc-fixed-medium-r-*-*-*-120-*-*-*-*-*-*");
     // "-bitstream-bitstream vera sans mono-medium-r-normal--15-140-0-0-m-0-iso8859-1");
#endif

    if(argc < 2) {
        *exename = 0;
        filename = exename;
        select_file("select or type the name of the command-line program to load");
        if(!*exename) return(0);
    } else {
        STRCPY(exename, argv[1]);
    }

    loadfile();

    startwindow();
    free_all();
    return(0);
}


